
var canvas, ctx, lastTime, rightClick;
canvas = document.querySelector('canvas');
canvas.width = window.innerWidth;
canvas.height = window.innerHeight;

var clothW = 47;
var clothH = 10;

var renderer = new ClothDemo.GLRenderer(canvas, clothW, clothH);

var prevMouse = null;
var running = false;
var startTime;
var hasTracked = false;

window.onmousemove = function(e) {
  e.preventDefault();

  var mouse = [e.pageX / renderer.scale,
               e.pageY / renderer.scale];

  if(prevMouse) {
    var diff = [mouse[0] - prevMouse[0], mouse[1] - prevMouse[1]];
    var d = Math.sqrt(diff[0] * diff[0] + diff[1] * diff[1]);

    for(var i=0; i<d; i+=1) {
      mousemove(prevMouse[0] + diff[0] * (i / d),
                prevMouse[1] + diff[1] * (i / d));
    }
  }

  mousemove(mouse[0], mouse[1]);
  prevMouse = mouse;
};

requestAnimationFrame(function() {
  init();
  start();
});

function init() {
  // responsiveness
  var scale = Math.min(canvas.width / 1359, 1.0);
  var v = [Math.max((canvas.width - 1359) / 2, 0),
           (1 - scale) * 50,
           Math.min(canvas.width / 1359, 1.0)];
  console.log(v);
  renderer.setOffset(v[0], v[1], v[2]);

  initPoints();
}

function start() {
  lastTime = Date.now();
  startTime = Date.now();
  running = true;
  requestAnimationFrame(heartbeat);
}

function stop() {
  running = false;
}

function heartbeat() {
  if(!running) {
    return;
  }

  var now = Date.now();
  var dt = now - lastTime;
  lastTime = now;

  update(16 / 1000);

  renderer.render(entities, clothW, clothH);
  //requestAnimationFrame(heartbeat);
}

function update(dt) {
  for(var z=0; z<3; z++) {
    for(var i=0; i<entities.length; i++) {
      entities[i].solveConstraints();
    }
  }

  for(var i=0, l=entities.length; i<l; i++) {
    entities[i].update(dt);
  }
}

var mouseInfluenceSize = 10;
var mouseInfluenceScalar = 8;
var lastMouse = [0, 0];
function mousemove(x, y) {
  if(rightClick) {
    for(var i=0; i<entities.length; i++) {
      if(entities[i].pinned) {
        continue;
      }

      var pos = entities[i].pos;
      var size = entities[i].size;

      if(x > pos[0] && x < pos[0] + size[0] &&
         y > pos[1] && y < pos[1] + size[1]) {
        entities[i].removeLinks();
        entities.splice(i, 1);
        break;
      }
    }
  }
  else {
    for(var i=0; i<entities.length; i++) {
      if(entities[i].pinned) {
        continue;
      }

      var pos = entities[i].pos;
      var line = [pos[0] - x, pos[1] - y];
      var dist = Math.sqrt(line[0]*line[0] + line[1]*line[1]);

      if(dist < mouseInfluenceSize) {
        renderer.fadeIn();

        entities[i].lastPos[0] =
          (entities[i].pos[0] -
           (x - lastMouse[0]) * mouseInfluenceScalar);

        entities[i].lastPos[1] =
          (entities[i].pos[1] -
           (y - lastMouse[1]) * mouseInfluenceScalar);
      }
    }
  }

  lastMouse = [x, y];
}

// objects

function Point(pos, size, mass, pinned) {
  this.pos = pos;
  this.lastPos = pos;
  this.size = size;
  this.mass = 1;
  this.acc = [0, .05];
  this.pinned = pinned;
  this.links = [];
}

Point.prototype.update = function(dt) {
  this.applyForce([0, this.mass * gravity]);
  var dtSeq = dt * dt;

  var x = this.pos[0];
  var y = this.pos[1];
  var lx = this.lastPos[0];
  var ly = this.lastPos[1];
  var acc = this.acc;

  var vel = [(x - lx) * .99, (y - ly) * .99];
  var next = [x + vel[0] + acc[0] * dtSeq,
              y + vel[1] + acc[1] * dtSeq];
  this.lastPos = this.pos;
  this.pos = next;

  this.acc = [0, 0];
};

Point.prototype.solveConstraints = function() {
  var links = this.links;

  for(var i=0; i<links.length; i++) {
    links[i].solve();
  }

  if(this.pinned) {
    this.pos[0] = this.lastPos[0];
    this.pos[1] = this.lastPos[1];
  }
};

Point.prototype.applyForce = function(force) {
  this.acc[0] += force[0] / this.mass;
  this.acc[1] += force[1] / this.mass;
};

Point.prototype.attachTo = function(point, distRest, stiffness, tearness) {
  this.links.push(new Link(this, point, distRest, stiffness, tearness));
};

Point.prototype.removeLink = function(link) {
  this.links.splice(this.links.indexOf(link), 1);
};

Point.prototype.removeLinks = function() {
  this.links = [];
};

function Link(point1, point2, distRest, stiffness, tearness) {
  this.point1 = point1;
  this.point2 = point2;
  this.distRest = distRest;
  this.stiffness = stiffness;
  this.tearness = tearness;
}

Link.prototype.solve = function() {
  var p1 = this.point1.pos;
  var p2 = this.point2.pos;

  var diff = [p1[0] - p2[0], p1[1] - p2[1]];
  var d = Math.sqrt(diff[0] * diff[0] + diff[1] * diff[1]);

  if(d > this.tearness) {
    this.point1.removeLink(this);
  }
  else if(d < this.distRest) {
    return;
  }

  var scalar = (this.distRest - d) / d;

  var im1 = 1 / this.point1.mass;
  var im2 = 1 / this.point2.mass;
  var scalarP1 = (im1 / (im1 + im2)) * this.stiffness;
  var scalarP2 = this.stiffness - scalarP1;

  if(!this.point1.pinned) {
    p1[0] += diff[0] * scalarP1 * scalar;
    p1[1] += diff[1] * scalarP1 * scalar;
  }

  if(!this.point2.pinned) {
    p2[0] -= diff[0] * scalarP2 * scalar;
    p2[1] -= diff[1] * scalarP2 * scalar;
  }
};

// state

var entities = [];
var gravity = 0;

var startY = 0;
var startX = 0;
var restingDistances = 30;
var stiffness = 1;
var curtainTearSensitivity = 70;

// the initial positions were made by creating a flat cloth
// and then interacting with it and calling
// `ClothDemo.savePoints` to grab a snapshot of the existing
// structure
var positions = [176.07970014172156, 117.56090037334924, 193.2109537527819, 104.55803242279038, 220.95966268164275, 115.960188040797, 233.19419498760533, 88.56828742666842, 260.8176859692299, 100.27054131109111, 288.87019864150204, 110.90334586918328, 261.8501711077318, 123.93861004253536, 268.3523702495211, 153.22549117249721, 295.4502112722858, 140.35276761046768, 322.097634995333, 126.5718094100822, 325.2039846753154, 96.73306627861619, 318.636774896099, 87.64669297023494, 339.44745321468935, 106.04478501152968, 357.04347735725304, 122.44986360486101, 375.8473530998583, 128.6615503445264, 396.30938601518676, 120.27591602824195, 425.9088066418599, 115.38978273439251, 453.3028279629242, 127.6195658829449, 483.2790101955437, 126.42436767505461, 512.6230298640136, 120.1850961095241, 541.5189346242375, 112.12118340612442, 571.4606865910638, 110.25263071719496, 596.7423974307711, 126.40302071302574, 619.9268509044944, 145.44195758196636, 649.5886252129013, 140.94983727232196, 678.9307270117617, 134.70155254567263, 708.777115896716, 131.66954487298332, 737.4458220700614, 122.83182535117692, 765.2734958940574, 111.62376710974479, 794.94957594673, 116.02038771028533, 817.7651379380488, 135.49986745222003, 847.7650890361182, 135.4457000039548, 875.9043612445726, 145.84672685607092, 905.3324407671976, 140.0167924136719, 932.1220349620082, 153.51929747271188, 959.4291901359143, 165.94182846714415, 989.3414307680109, 168.23483094512366, 1017.9132829890644, 159.08883397333557, 1044.244008377189, 144.71201581210545, 1073.9725091900846, 148.73895502416266, 1095.9412328151586, 169.16871719324152, 1124.8847887852519, 161.27754695691004, 1137.4514311167786, 134.03641077444948, 1152.2287329836872, 108.29808872674023, 1178.5050463941534, 93.82206187179776, 1174.6484997971968, 67.14578091843102, 1203.6857651859043, 74.68482719337673, 165.54240688339547, 145.48567327500655, 190.94146654929517, 134.32428415944165, 218.15853906327735, 136.72826787960412, 237.84692454690907, 114.09273216606167, 245.5977075836767, 85.11126632471364, 272.9526780475274, 93.99482545799354, 283.3843495893561, 122.122751595877, 275.7297478798789, 145.54566471920106, 295.5264229906817, 123.00477621589557, 311.44719123264247, 101.44107974731028, 325.4426505575526, 86.61143269278286, 332.0301960716699, 114.49098974263745, 360.4271519653154, 124.16636549625477, 386.9845294769166, 120.57013012200235, 403.8825954940316, 118.26033118471808, 396.991862808126, 144.92131667831939, 424.35913604841306, 139.89290353530208, 450.2185160647092, 124.6846010287978, 480.09668047656265, 121.98362132269571, 510.06486820581597, 123.36482522524204, 539.9883730587711, 121.22383076138262, 569.703339043778, 117.09820395991511, 599.191477083631, 111.58007792501927, 621.555879120069, 131.5759155306185, 648.7460354921711, 144.25248432743612, 678.7375623894173, 144.96544601733444, 707.2650362564585, 135.68195342223308, 733.5740263363622, 121.26539901514262, 763.0791151019826, 115.83863783281873, 791.9774573629514, 123.89381098805133, 818.5682904211083, 137.78364890172753, 847.1497593325882, 146.84231051281637, 876.897997923431, 150.72075101641886, 906.3105330026602, 156.62860851411708, 935.5546856644398, 163.3203597854974, 965.1892058365752, 158.6518253085017, 993.6732290599662, 149.2358590277956, 1022.2270593122046, 140.03375177328937, 1051.49948401227, 133.4667774261003, 1081.496688404782, 133.05723057891794, 1089.8639292732928, 156.91414672316682, 1103.9596097810513, 143.5267062125413, 1133.4611881856576, 138.08089393291007, 1122.7568191602963, 113.90222072864738, 1149.0633427005334, 99.58099689578343, 1153.6865053403853, 69.94105227267607, 1182.2312226175773, 60.71071557445865, 140.69224895008205, 128.67883982291568, 165.21819483159493, 138.69183471152536, 191.36337821698714, 123.98029550775018, 221.34210304249055, 122.85066820782176, 236.72685715286252, 97.0958756254553, 247.31071783344768, 100.72317260248855, 260.40493537268543, 127.71467991562535, 282.770646861992, 147.70905286346002, 312.71982412762515, 145.96354978432012, 303.2986002706974, 119.91640904031182, 332.99306801616547, 115.64574336482528, 356.5079907613373, 130.09036618999846, 384.9539265519072, 120.55995502387233, 388.63823464235577, 90.78704999438865, 418.4981040965138, 93.19489068803985, 413.09791402575945, 119.63162524288066, 435.54923972311855, 139.52981547259088, 465.2685428092317, 143.62408269343177, 493.0293208904005, 132.2513433432167, 512.543952955006, 109.46584059820788, 540.6704968204687, 119.90088912024218, 570.2430009309383, 114.85495599461261, 599.5891061916525, 121.08427999306898, 629.0400933293929, 126.79736642769319, 658.8243288893439, 123.32765276748145, 688.7013667986482, 121.43320559981902, 718.6474406050922, 120.80650599081733, 748.0831874084868, 126.5096840818075, 778.0666084146793, 127.5069126212786, 806.8706576825037, 119.12080829739213, 836.0617325827011, 113.41195463701716, 854.3498170370419, 137.0554504229569, 875.0243033974443, 158.79402474893698, 904.9812513896061, 157.187392559869, 932.7864721069363, 145.9237474640699, 962.6283202146034, 142.84736947754007, 990.0544785150714, 155.00491202847, 1013.1946575713057, 164.6070818734384, 1042.808057009989, 159.8063971641265, 1070.821151233731, 159.2948813208455, 1070.2670675875509, 134.1993275059705, 1081.7829346160004, 130.84953828057263, 1110.3599555434134, 121.71970396129281, 1100.7695345781765, 93.49243590114669, 1127.896176356891, 94.5603949892741, 1150.0788059083366, 74.36309452185039, 1164.9649053237986, 48.38521648034728, 111.83285593718034, 120.88041969228682, 139.1856298590515, 123.78191985789671, 166.08959248963532, 110.50875916527549, 195.8827484664399, 106.99196276204216, 212.79474342185378, 98.72029360597213, 228.6360592537558, 124.19680013274132, 255.56624047987512, 131.7291691063613, 269.570149097122, 158.26010469934636, 298.7217197059519, 151.1759017774982, 323.9759276867838, 134.98253958026618, 353.91029868566034, 132.99925117435527, 372.97053458188054, 109.8323047295726, 393.4977850770185, 92.24767515987008, 400.80915663097403, 63.36682647329077, 428.45633296563784, 68.9533867290467, 434.2761139098314, 98.38347591950028, 426.15355319143555, 121.28743123169309, 450.4144503104024, 138.9342119852455, 480.05176601297387, 134.2834574288114, 509.83124680054505, 130.65268285328827, 539.8093551369294, 131.79855393824823, 567.4103158188346, 120.04325855974226, 595.7636128372606, 110.24067991042901, 625.7587264278302, 109.69923675716028, 655.7412900172424, 110.72191980022123, 678.5779975407041, 130.176605340465, 706.4329353322961, 119.0364778444915, 734.7082198452547, 109.01109577171653, 764.5124726196758, 105.58961706251426, 794.3057279971227, 109.10557127298442, 824.2748610755285, 107.7450336849385, 850.5939173357045, 122.14320292867275, 878.1307217907586, 134.04800869718264, 906.2709397177198, 144.4464766138048, 931.4680470415749, 160.72854596767334, 960.6405638626104, 156.65973510194894, 990.1268346093353, 159.04526463507264, 1020.0567455912327, 156.99576991312694, 1049.384245101509, 150.67930155878088, 1072.9707623606012, 132.14079558917567, 1050.54167697527, 121.05712567173506, 1068.7331917332876, 144.0595578608873, 1088.5631969501835, 122.72210280462959, 1096.983374570103, 107.25279716631822, 1102.0010466832678, 79.41304343331089, 1124.773152946022, 69.05053746396791, 1135.617426807604, 42.162234500170605, 126.55722774949392, 94.74246118875504, 147.7829292274057, 97.44307257276066, 171.80660442490648, 113.11059629561309, 201.79524446289585, 113.93610745318922, 229.59671238487857, 123.50097353924376, 247.53620326776476, 147.49453728132343, 271.2470316823708, 157.02108343543568, 293.8799468334841, 175.8276017878609, 314.70455940779533, 154.2327989147197, 338.3344273165975, 135.74958105059866, 368.2985650936724, 134.28314070866114, 394.3026835406336, 119.32366799749254, 415.6985653199742, 109.98354225742763, 398.7893538839304, 91.79680971077597, 418.877718575235, 83.30532407694774, 444.8205139330535, 97.59142686584539, 425.0536637431123, 117.98313386860097, 451.8122138234937, 131.5470565164, 480.61858513485026, 123.16893191332665, 508.0001537697374, 110.91129351559742, 537.9415705676813, 109.0373778009666, 567.9070654221016, 110.47582075378804, 597.8234685461354, 108.23778051382885, 627.5111557727464, 112.55532925364182, 657.4929197203586, 111.50946569443335, 687.1861488913381, 107.23019690263544, 717.1835909647375, 107.62194816412443, 747.1463726453458, 108.53626514037255, 777.1021656134025, 106.91587328192304, 794.7765675981908, 110.96080475706835, 818.1696675707701, 122.03395889914367, 846.2020951260196, 130.29316981889062, 867.6319458573214, 136.14592332161135, 885.1746189574121, 133.46177266510753, 913.9134829890486, 142.00559378544142, 942.4637764060366, 132.79334874461068, 972.0955653142105, 135.15949517495855, 1002.094846868399, 135.36711556016618, 1028.2045351093423, 130.01066292611475, 1055.7004786073965, 118.01178418059193, 1032.233684444863, 102.49167851850112, 1058.037049656435, 116.44602512899522, 1072.2207282010452, 142.881290645785, 1067.690312073099, 113.72709167803445, 1075.6751611709078, 93.79872214321206, 1096.635649095079, 79.45634723593948, 1115.8202245390328, 64.70266001363119, 155.5876629048413, 102.30776585483082, 172.34362198718966, 114.56593845310525, 198.96975481441507, 118.76804616934183, 219.4485372241068, 138.1922665614849, 248.14246006144663, 146.94776724410895, 277.24340658610447, 143.3133806235661, 286.4288111240828, 171.87258830630714, 316.0737671856056, 167.27078647345883, 342.8692484655303, 159.5137151010075, 355.78638006518446, 133.84934881562089, 361.4784594362775, 106.88071130779636, 385.2674061124108, 90.71661009586106, 401.02003219884625, 100.99195779993413, 371.0258369704412, 101.58208710761636, 394.1312032570011, 100.14999178201583, 420.2824574085606, 114.85073687513595, 428.7828381957879, 143.6212691341904, 450.6465719078863, 127.35755873075306, 475.05703230540547, 109.91825188730475, 500.0924144538414, 94.83235498516188, 529.118684435559, 87.63729069304684, 557.3986126732274, 90.63124206774872, 587.3014728230961, 92.9716394083248, 617.1314851369611, 96.14886698579255, 641.5912623250795, 113.51631320471178, 671.2687682215349, 117.8932510603155, 700.9214356068558, 122.43517320267809, 729.1646477267946, 132.54996058594972, 758.5581989388984, 126.79255338894633, 782.6506410677922, 138.40095610098928, 795.3217523524155, 131.13960769363027, 825.2443405719487, 128.98584024084315, 845.8901210195984, 138.14556385379498, 868.2036855826267, 158.19811572360385, 896.2678154045522, 166.26730055754246, 915.6708338099277, 146.0060799527939, 943.5908742692848, 139.8563287340231, 972.4799967934406, 131.86662742401, 999.8534395941019, 120.20171889123834, 1028.3275572430755, 126.67664143545336, 1045.7362690727323, 102.30907075982664, 1034.6307134261542, 116.02476432207376, 1048.3846455489404, 125.28686926180552, 1041.7970305049244, 98.67272739092687, 1055.258026169351, 71.86225190613098, 1077.1445521007934, 62.3438033284247, 1088.3357080676185, 76.72769033914172, 181.57856915811206, 117.29018195283233, 199.13200951492666, 114.7857942976115, 218.27856835740852, 137.88144939746454, 218.09544884030421, 167.18837270160486, 246.02053519747633, 156.22527314250988, 275.96758417979385, 158.00691767724368, 305.9240174213962, 156.3907161301038, 331.8485138397383, 141.7530011464271, 354.7052739276535, 131.94727715084738, 376.60628492731206, 112.25000712358168, 382.0365580352077, 85.75125293487865, 391.23372338512166, 61.31587782273433, 403.5496086871963, 71.09879398630237, 386.1461771630761, 83.52022224295229, 386.22187812012265, 113.51102672401518, 414.01670533843105, 124.78115897825386, 435.09377377011106, 137.9836889131231, 458.8599289499585, 120.85101262901111, 487.3585952678224, 111.47945896369947, 516.4513200857571, 110.54565984751633, 546.4081713855722, 112.15409393093337, 574.3485285242459, 113.12676706190308, 602.9066076902427, 112.91785337071353, 632.9012744363378, 113.4835094402792, 662.85147848344, 111.7557138689427, 683.5105734472436, 133.50891582350536, 705.4987601440154, 151.91843480823388, 734.637703953463, 159.05439680849975, 763.23968001272, 150.00304522419177, 776.7728883843766, 123.31375401864568, 798.29900286269, 102.41814693329809, 828.1626448949995, 105.1773706480519, 857.6671833142079, 110.5935630733478, 854.1847809612323, 140.27818909416305, 866.2806286909931, 167.14401773317994, 896.2249529792958, 168.85028531695124, 915.6972398037019, 148.65756057084818, 942.7155641183203, 135.61876648435336, 970.0551832049168, 123.63556945205521, 999.1190518599508, 119.83097030090643, 1023.4285331185458, 122.36249537963441, 1046.1455215467931, 102.76814831854006, 1024.925170308852, 106.61798655004306, 1022.5889393829383, 77.7631266494308, 1034.4435392821638, 50.25768906373383, 1049.2631857248443, 51.26998695838691, 1058.36329260217, 75.44149005822013, 211.0595909140384, 122.84620177153411, 197.59956825720735, 134.70153636760722, 207.42448975447672, 155.98501245168114, 236.51405465500127, 148.66248317965096, 266.4985571875422, 149.62664492714492, 287.6599802936013, 130.37925268150565, 311.95778790464806, 127.00375280018207, 335.0736120884215, 111.926859235592, 365.0626535193053, 111.11606083120863, 374.4554793060706, 82.62439847643074, 377.24327411410434, 56.136654689267665, 400.0520827042807, 36.6492677251694, 423.4950443052666, 48.68942991676545, 409.53524524583884, 64.85613535518776, 409.71550108581005, 94.85515921493285, 436.36127492887414, 108.63915964461022, 455.79925043694925, 116.27463006666434, 484.4726863912027, 125.09699202796826, 514.4519036761254, 123.98051050851961, 543.8726666097593, 118.1137645277841, 570.9815778889206, 129.36303616515522, 596.5043944388976, 113.59643109794958, 621.7646865593005, 129.78030090685965, 651.3038399200798, 135.01847031299266, 681.2694050425252, 133.58149192430758, 708.204649141458, 146.7910575996358, 734.0806546390648, 156.95099393050452, 762.9095094794692, 149.51937183749862, 777.6305314065723, 123.67998696417848, 773.060846195251, 94.03006323629293, 795.1198312908776, 73.69779389546359, 825.0750318888702, 75.33668288915402, 844.8643770377416, 83.46261853676938, 857.8444669494784, 110.50224748911975, 853.3217402333634, 140.08995459573742, 866.618758180194, 166.9821340113312, 894.9713038642376, 157.17738244732885, 913.7546824922728, 133.78540332382158, 940.2901289241217, 119.98101802879819, 969.2593536636757, 116.93296691527213, 996.1663717228151, 109.84153071903124, 1024.5611691884076, 117.51909163615353, 1048.7765319342222, 99.80987914420604, 1043.9058666652143, 70.20790928006934, 1049.2345212595546, 40.6849446556586, 1030.9782914520767, 60.98099494423437, 1029.8644138356299, 66.07058246903145, 220.73766061781555, 151.24223964318952, 203.90690914813084, 164.03100023729667, 226.02518208858257, 179.5225196410343, 234.94694037086563, 150.88439766937623, 250.09508490162656, 124.98973186202961, 272.60661863743644, 105.1596827228962, 284.9312308414012, 113.98203197281744, 309.32913015978784, 96.52519616983233, 339.190304797828, 99.40794646007986, 358.27291489779157, 76.25942616676835, 381.83229578501454, 57.81461874662766, 400.0734193339647, 36.34577141625157, 396.2046622731518, 43.610658049604126, 397.952255890145, 73.55971340281687, 418.13228176223646, 95.75805919257213, 447.7076318799079, 90.72828132308355, 476.72949094719974, 98.32641900795363, 503.3297948778095, 112.1981107144176, 530.900143287964, 124.02502445142585, 558.2530485542479, 136.34649244403894, 586.4683434707615, 126.1534981453566, 616.4664244935036, 126.49281356150527, 646.2743593606248, 129.88206406213587, 674.7694813622629, 139.2643890464438, 702.893001403874, 149.70793348625133, 714.4388294397436, 176.136159337548, 744.3389256511034, 178.58243130740587, 745.4999481013236, 148.60490594473663, 749.0883264282393, 118.8202867431618, 768.2205569687736, 95.83430436678061, 790.7884623049138, 76.06843356036644, 804.1668717757359, 65.35326912265191, 816.7648248221441, 92.5799394853625, 835.3625020923863, 116.11982903801275, 845.6779907190286, 144.29057085276176, 844.6111194401407, 174.271594612371, 874.5683485727589, 175.8729761083624, 889.2015638660122, 149.6838751276242, 912.1397308420993, 130.35189493571687, 942.0386007948154, 129.54370096416582, 967.119100644929, 113.24196097451174, 996.5088355608863, 107.22170503575651, 1026.4642958825525, 108.85583994931811, 1035.1699549338, 80.14675601265434, 1025.2878689772037, 51.8210719583249, 1001.2054036342149, 64.66544210598826, 1000.7802754755517, 73.42672926951784, 197.99104090598922, 159.08882143754798, 187.63030043965128, 187.24295180927749, 213.18470340169412, 202.95831019695947, 233.41775090538295, 180.80828110665092, 242.0612844929094, 153.09948217759077, 254.1981680028168, 125.66417538749255, 261.48089724137145, 96.56156631480557, 285.57768202251526, 78.69134070879241, 309.9307847167953, 96.21065545454442, 339.75421966235115, 99.46069053845899, 364.222168868065, 82.10213467495257, 383.4522678845528, 61.32050513846979, 409.62179761702004, 51.643105565828925, 410.99476801472053, 81.61167163573529, 415.3647783469918, 110.79240237110733, 443.35467395892806, 100.03682973984321, 472.30060025547334, 107.8533047349601, 502.29367374511787, 107.20867876398918, 531.5907598049074, 113.57239174803186, 560.2140041001089, 122.47914258261719, 589.776430871241, 127.47997579095266, 619.6928706535259, 129.71752595826837, 649.6330528858264, 127.82398749925208, 679.4363260430956, 131.25398882756696, 695.1355343280145, 156.818316693298, 705.4703802495869, 184.98196276012192, 735.2316631934068, 181.20492191608676, 727.474875444537, 152.22506263729127, 735.9788833338769, 123.45560227012538, 754.0873291760448, 99.53732078544748, 783.9239865809453, 96.4110010799143, 811.9718246580052, 85.76587171466355, 840.0020671793562, 96.45724693921399, 853.5971899040437, 123.19995880806688, 849.3052233825535, 152.89135526189523, 843.206577451528, 182.26492361522713, 856.7447411794212, 165.514241383384, 877.379978704417, 143.73840678895593, 907.377544948157, 144.1205317055544, 930.5235463268181, 125.03486639943071, 953.5069501021785, 105.75370588200647, 982.1474976628924, 114.68225757990537, 1011.461734318285, 121.05999458539696, 1024.902277369761, 94.23926002272435, 1007.0135193385902, 70.15622997868391, 977.4187981732781, 75.0707471621091, 990.995365544833, 101.7223197270627];

function initPoints() {
  entities = [];

  for(var y=0; y<clothH; y++) {
    for(var x=0; x<clothW; x++) {
      var p = new Point(
        // [startX + x * restingDistances,
        //  y * restingDistances + startY],
        [positions[(y * clothW + x) * 2],
         positions[(y * clothW + x) * 2 + 1]],
        [3, 3],
        1,
        false
      );
      p.scale = Math.random() * 10 | 0;

      if(x > 0) {
        p.attachTo(entities[entities.length - 1],
                   restingDistances,
                   1,
                   curtainTearSensitivity);
      }

      if(y > 0) {
        p.attachTo(entities[(y - 1) * clothW + x],
                   restingDistances,
                   1,
                   curtainTearSensitivity);
      }

      entities.push(p);
    }
  }
}

window.ClothDemo.savePoints = function() {
  var points = [];

  for(var i=0, l=entities.length; i<l; i++) {
    points.push(entities[i].pos[0]);
    points.push(entities[i].pos[1]);
  }

  console.log(points.toSource());
};

